/* -*- mode: c++; c-basic-offset: 4; indent-tabs-mode: nil -*-

this file is part of rcssserver3D
Fri May 9 2003
Copyright (C) 2002,2003 Koblenz University
Copyright (C) 2003 RoboCup Soccer Server 3D Maintenance Group

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef AGENT3D_WORLDMODEL_H
#define AGENT3D_WORLDMODEL_H

#include <string>
#include <map>
#include <zeitgeist/leaf.h>
#include <soccer/soccertypes.h>
#include <oxygen/gamecontrolserver/predicate.h>
#include <oxygen/gamecontrolserver/baseparser.h>

class WorldModel
    : public zeitgeist::Leaf {
public:
    struct VisionSense {
        /** distance perceptor to object */
        float distance_;

        /** theta is the angle in the X-Y (horizontal) plane */
        float theta_;

        /** phi is the latitude angle */
        float phi_;

        VisionSense()
            : distance_( 0 )
            , theta_( 0 )
            , phi_( 0 )
          { }
    };

    enum VisionObject {
        VO_BALL   = 0,
        VO_FLAG1L = 1,
        VO_FLAG1R = 2,
        VO_FLAG2L = 3,
        VO_FLAG2R = 4,
        VO_GOAL1L = 5,
        VO_GOAL1R = 6,
        VO_GOAL2L = 7,
        VO_GOAL2R = 8,

        VO_OUR0,
        VO_OUR1,
        VO_OUR2,
        VO_OUR3,
        VO_OUR4,
        VO_OUR5,
        VO_OUR6,
        VO_OUR7,
        VO_OUR8,
        VO_OUR9,
        VO_OUR10,
        VO_OUR11,

        VO_OPP0,
        VO_OPP1,
        VO_OPP2,
        VO_OPP3,
        VO_OPP4,
        VO_OPP5,
        VO_OPP6,
        VO_OPP7,
        VO_OPP8,
        VO_OPP9,
        VO_OPP10,
        VO_OPP11,
    };

    enum PlayersSet {
        SET_ALL,
        SET_TEAMMATE,
        SET_TEAMMATE_NOGK,
        SET_OPPONENT,
    };

    enum FormationType {
        FT_ERROR = 0,
        FT_initial,
        FT_433_offensive,
        FT_MAX
    };

    //! formation parameter
    class FormationParam {
    private:
        //! base point
        salt::Vector3f M_pos;
        //! attraction param
        salt::Vector3f M_attract;
        float M_min_x;
        float M_max_x;
        float M_min_y;
        float M_max_y;
        bool M_behind_ball;
        bool M_have_etc;
    public:
        FormationParam()
            : M_have_etc( false )
          { }
        ~FormationParam()
          { }
        void setValues( float px, float py,
                        float ax = 0.0, float ay = 0.0,
                        float minx = -53, float maxx = 53,
                        float miny = -30, float maxy = 30,
                        bool behind_ball = false )
          {
              this->M_pos  = salt::Vector3f( px, py, 0 );
              this->M_attract = salt::Vector3f( ax, ay, 0 );
              this->M_min_x = minx;
              this->M_max_x = maxx;
              this->M_min_y = miny;
              this->M_max_y = maxy;
              this->M_behind_ball = behind_ball;
              M_have_etc = true;
          }
        const
        salt::Vector3f & pos() const
          {
              return M_pos;
          }
        const
        salt::Vector3f & getAttract() const
          {
              return M_attract;
          }
        float getMinX() const
          {
              return M_min_x;
          }
        float getMaxX() const
          {
              return M_max_x;
          }
        float getMinY() const
          {
              return M_min_y;
          }
        float getMaxY() const
          {
              return M_max_y;
          }
        bool isBehindBall() const
          {
              return M_behind_ball;
          }
    };

    class Formation {
    private:
        //FormationParam M_params[12];
    public:
        FormationParam player[12];
        Formation()
          { }
        ~Formation()
          { }
    };

    Formation M_formations[FT_MAX];

protected:
    // mapping from gamestate string to TPlayMode
    typedef std::map< std::string, TPlayMode > TPlayModeMap;
    TPlayModeMap M_playmode_map;

    // mapping from vision object to VisionSense
    typedef std::map< VisionObject, VisionSense > TVisionMap;
    TVisionMap M_vision_map;

    // mapping from <ObjectName>+<ID> to VisionObject
    typedef std::map< std::string, VisionObject > TVisionObjectMap;
    TVisionObjectMap M_vision_object_map;

    // parser instance
    boost::shared_ptr< oxygen::BaseParser > M_parser_ptr;

    // mapping from vision object to object position
    typedef std::map< VisionObject, salt::Vector3f > TPositionMap;
    TPositionMap M_position_map;

    // soccer variables
    float M_field_length;
    float M_field_width;
    float M_field_height;
    float M_goal_width;
    float M_goal_depth;
    float M_goal_height;
    float M_border_size;

    // agent parameter
    float M_agent_mass;
    float M_agent_radius;
    float M_agent_max_speed;

    // ball parameter
    float M_ball_radius;
    float M_ball_mass;

    // team information
    TTeamIndex M_team_index;
    unsigned int M_team_unum;
    std::string M_team_name;
    std::string M_opp_team_name;

    // game state information
    float M_time;
    TPlayMode M_playmode;

    // the players position
    salt::Vector3f M_my_pos;

    // the players position from Vision
    salt::Vector3f M_my_pos_v;
    bool M_my_pos_v_enable;

public:
    WorldModel();
    virtual
    ~WorldModel();

    /** parses a message received from the server */
    void parse( const std::string& message );

    /** returns the polar coordinates of the requested object */
    VisionSense getVisionSense( VisionObject obj );

    /** returns the global cartesian coordinates of the sensed
        object */
    salt::Vector3f getPosition( VisionSense sense );

    /** returns the coordinates of the requested object */
    salt::Vector3f getObjectPosition( VisionObject obj );

    /** returns the current PlayMode */
    TPlayMode getPlayMode() const
      {
          return M_playmode;
      }

    /** given an object position in polar coordinates and distance,
        CalcDriveVec returns a cartesian normalized vector that can be
        passed to the drive effector to reach this position.
    */
    salt::Vector3f getDriveVec( const WorldModel::VisionSense& vision );

    /** returns the players calculated position */
    const
    salt::Vector3f & getMyPosition() const
      {
          return M_my_pos;
      }

    /** returns the players team index */
    TTeamIndex getMyTeam() const
      {
          return M_team_index;
      }


    float getFieldWidth() const
      {
          return M_field_width;
      }
    float getFieldLength() const
      {
          return M_field_length;
      }
    float getGoalWidth() const
      {
          return M_goal_width;
      }
    float getGoalDepth() const
      {
          return M_goal_depth;
      }
    float getGoalHeight() const
      {
          return M_goal_height;
      }

    float getAgentRadius() const
      {
          return M_agent_radius;
      }
    float getBallRadius() const
      {
          return M_ball_radius;
      }

    float getKickableDistance() const
      {
          return M_agent_radius + M_ball_radius + 0.04;
      }
    int getTeamUnum()
      {
          return M_team_unum;
      }

    void setTeamName( const std::string& name )
      {
          M_team_name = name;
      }

    void setOppTeamName( const std::string& name )
      {
          M_opp_team_name = name;
      }

    const
    std::string & getTeamName() const
      {
          return M_team_name;
      }

    const
    std::string & getOppTeamName() const
      {
          return M_opp_team_name;
      }

    VisionObject getFastestInSetTo( PlayersSet ps, VisionObject vo );
    salt::Vector3f getStrategicPosition();

    salt::Vector3f getObjectSpeed( VisionObject obj );
    VisionObject getMyObject();

    salt::Vector3f getFromMyPosTo( salt::Vector3f pos );
    int WorldModel::getNeedStepsSetTo( VisionObject po, VisionObject vo );

protected:
    virtual
    bool ConstructInternal();

    void calcPlayerPosition();

    void parseObjectVision( const oxygen::Predicate& predicate );
    void parseGameState( const oxygen::Predicate& predicate );
    void parseVision( const oxygen::Predicate& predicate );

    void parsePlayMode( const oxygen::Predicate& predicate );
    void parseTeamIndex( const oxygen::Predicate& predicate );

    void setupGameStateMap();
    void setupVisionObjectMap();
};

DECLARE_CLASS(WorldModel);

#endif // WORLDMODEL_H
